/*
 * comm.c
 *
 * handle communications with a pilot
 *
 * Copyright (C) 1997 by David Lindes,
 * see the LICENSE file in this directory for licensing information
 *
 * $Id: comm.c,v 1.5 1997/04/23 03:18:47 lindes Exp $
 *
 */

#include	<pwd.h>
#include	<stdio.h>
#include	<stdarg.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<sys/stat.h>

#include	"pi-task.h"
#include	"pi-ui.h"
#include	"util.h"

void	begin_talking(PilotConnection *pilot);
void	identify_user(PilotConnection *pilot);
void	setup_save(PilotConnection *pilot);
void	get_sys_info(PilotConnection *pilot);
void	disconnect_from(PilotConnection *pilot);
void	synchronize_clock(PilotConnection *pilot);

extern int	errno;

void	begin_talking(PilotConnection *pilot)
{
    dlp_OpenConduit(pilot->connectionfd);
}

void	identify_user(PilotConnection *pilot)
{
    struct	passwd	*pwinfo;
    char		*realname, *tmp;

    send_message("Identifying User...\n");

    dlp_ReadUserInfo(pilot->connectionfd,&(pilot->user));

    if(!(pilot->user.username[0]))
    {
	/* ugly, but it's something... */
	send_message("\n\nUninitialized pilot! Getting username...\n");
	strncpy(pilot->user.username,get_user_name(),127);
	pilot->user.username[127] = '\0'; /* assure termination */

	pilot->user.lastSyncDate = time( (time_t *)0);
	dlp_WriteUserInfo(pilot->connectionfd,&pilot->user);
    }

    setpwent();

    while((pwinfo = getpwent()) != NULL)
    {
	realname	= strdup(pwinfo->pw_gecos);
	verify_alloc(realname);

	tmp		= strchr(realname,',');
	if(tmp)
	    *tmp	= '\0';

	if(!strcmp(realname,pilot->user.username))
	{
	    pilot->uname	= strdup(pwinfo->pw_name);
	    verify_alloc(pilot->uname);

	    send_message("Openning connection for %s (%s)\n",
		pilot->uname, pilot->user.username);

	    free(realname);
	    endpwent();
	    return;
	}
	free(realname);
    }

    endpwent();

    /* couldn't find a passwd entry for this user... */
    send_message("Couldn't find an account for %s, using real name.\n",
	pilot->user.username);
    pilot->uname	= pilot->user.username;
}

void	setup_save(PilotConnection *pilot)
{
    struct stat	dirstat;
    ino_t	ino;
    FILE	*infofile;
    time_t	clock;

    send_message("Setting up SaveDir: %s/%s\n",
	pilot->saveroot, pilot->uname);

    if(stat(pilot->saveroot,&dirstat) == -1)
    {
	int	saveerrno;

	saveerrno	= errno;

	if(mkdir(pilot->saveroot,0777) == -1)
	{
	    if(errno != saveerrno)
	    {
		fprintf(stderr,"Couldn't create %s...\n",
			pilot->saveroot);
		perror("mkdir");
	    }
	    errno	= saveerrno;
	    fprintf(stderr,"Couldn't access %s...\n",
			pilot->saveroot);
	    perror("stat");
	    exit(1);
	}
	if(stat(pilot->saveroot,&dirstat) == -1)
	{
	    perror("Stat...");
	    exit(1);
	}
    }

    if(chdir(pilot->saveroot) == -1)
    {
	fprintf(stderr,"Couldn't change to %s:\n",
			pilot->saveroot);
	perror("chdir");
	exit(1);
    }

    if(stat(pilot->user.username, &dirstat) == -1)
    {
	int	saveerrno;

	saveerrno	= errno;
	if(mkdir(pilot->user.username,0777) == -1)
	{
	    if(errno != saveerrno)
	    {
		fprintf(stderr,"Couldn't create %s/%s...\n",
			pilot->saveroot,pilot->user.username);
		perror("mkdir");
	    }
	    errno	= saveerrno;
	    fprintf(stderr,"Couldn't access %s/%s...\n",
			pilot->saveroot,pilot->user.username);
	    perror("stat");
	    exit(1);
	}
	if(stat(pilot->user.username, &dirstat) == -1)
	{
	    perror("stat, second try");
	    exit(1);
	}
    }

    ino	= dirstat.st_ino;

    if(stat(pilot->uname, &dirstat) == -1)
    {
	if(symlink(pilot->user.username,pilot->uname) == -1)
	{
	    fprintf(stderr, "Couldn't create symlink %s/%s...\n",
			pilot->saveroot,pilot->uname);
	    perror("symlink");
	    exit(1);
	}
	if(stat(pilot->uname, &dirstat) == -1)
	{
	    perror("stat number 2, second try");
	    exit(1);
	}
    }

    if(dirstat.st_ino != ino)
    {
	fprintf(stderr,
		"Internal Error: %s != %s in %s.\n",
		    pilot->user.username,
		    pilot->uname,
		    pilot->saveroot
	    );
    }

    if(chdir(pilot->uname) == -1)
    {
	fprintf(stderr,"Couldn't change to %s/%s:\n",
		    pilot->saveroot, pilot->uname);
	perror("chdir");
	exit(1);
    }

    if(!(infofile = fopen(".laststart","w")))
    {
	perror("fopen");
	exit(1);
    }

    if(!(pilot->logfile = fopen("pilot-sync.log","a")))
    {
	perror("fopen");
	exit(1);
    }

    time(&clock);
    fprintf(pilot->logfile,"\nSynchronizing at %s",
		ctime(&clock));
    fprintf(infofile,"Running at                    %s",
		ctime(&clock));
    /*
    fprintf(infofile,"Previous Save at:             %s (%ld)\n",
		ctime(&(pilot->user.lastSyncDate)),
		pilot->user.lastSyncDate);
    */
    fprintf(infofile,"Previous Successfull Save at: %s",
		ctime(&(pilot->user.succSyncDate)));
    fclose(infofile);
}

void	get_sys_info(PilotConnection *pilot)
{
    send_message("Reading System Information\n");
    dlp_ReadSysInfo(pilot->connectionfd,&(pilot->system));

    fprintf(pilot->logfile,"SysInfo: ROM version: %ld\n",
		pilot->system.ROMVersion);
    fprintf(pilot->logfile,"SysInfo: Localization ID: %ld\n",
		pilot->system.localizationID);
    fprintf(pilot->logfile,"SysInfo: Name: %.*s\n",
		pilot->system.namelength, pilot->system.name);
    fflush(pilot->logfile);
}

void	disconnect_from(PilotConnection *pilot)
{
    dlp_EndOfSync(pilot->connectionfd,0);
    /* pi_close(pilot->connectionfd); */
    pilot->connectionfd	= -1;
    fclose(pilot->logfile);
}

void	synchronize_clock(PilotConnection *pilot)
{
    time_t	thetime;
    char	buf[BUFSIZ];

    if(pilot->setclock == 0)
    {
	return;
    }
    thetime	= time(NULL);
    send_message("Setting pilot clock to %s",ctime(&thetime));
    dlp_SetSysDateTime(pilot->connectionfd,thetime);
    sprintf(buf,"Set the pilot clock to %s",ctime(&thetime));
    log_message(pilot,buf);
}
