/*
 * tasks.c
 *
 * handle task registrations and callbacks
 *
 * Copyright (C) 1997 by David Lindes,
 * see the LICENSE file in this directory for licensing information
 *
 * $Id: tasks.c,v 1.9 1997/04/23 03:17:03 lindes Exp $
 *
 */

#include <dirent.h>
#include <dlfcn.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/stat.h>
#include <pi-file.h>
#include "pi-task.h"
#include "pi-ui.h"

void	get_registrations(PilotConnection *pilot);
int	registry_resize(void);
int	run_callback(PilotConnection *pilot);
void	backup_data(PilotConnection *pilot);

typedef struct cb_registration {
    unsigned long	type;
    unsigned long	creator;
    pi_cb		callback;
} cb_registration;

char	*taskdirname	= "pilot_tasks";
char	*progname;

cb_registration	*callbacks;
int		callback_count	= 0;
int		callback_size	= 0;

int		*verbosity	= NULL;

void	get_registrations(PilotConnection *pilot)
{
    char	*ldlpath, *token, *dirname = NULL;

    /* allow us to use verbosity as a global within this file,
       while still allowing possible updates from elsewhere.  */
    verbosity	= &(pilot->verbosity);

    /* figure out what directories to search */
    if(!(ldlpath = getenv("LD_LIBRARY_PATH")))
    {
	ldlpath	= pathadd(':',
		"/usr/local/lib/pilot",
		getenv("HOME"),
		NULL);
    }
    
    /* copy this information, as we will be mucking with it. */
    ldlpath	= strdup(ldlpath);

    if((*verbosity) > 0)
	fprintf(stderr,"Searching %s for tasks...\n",ldlpath);

    /* loop through the searchable directories */
    for(token=strtok(ldlpath,":");token;token=strtok(NULL,":"))
    {
	DIR		*conddir;
	struct dirent	*filep;

	if(dirname)
	    free(dirname);

	dirname	= pathadd('/',token,taskdirname,NULL);

	/* now open the directory... */
	conddir	= opendir(dirname);

	/* if it doesn't exist, that's fine, that just means we
	 * won't find any task libs there.  non-fatal. */
	if(!conddir)
	    continue;

	/* now start checking files... */
	while((filep = readdir(conddir)) != NULL)
	{
	    void	*library;
	    char	*filename = NULL;
	    int		(*run_registration)();

	    /* if it's not a .so file, we don't care about it */
	    if(strcmp(&filep->d_name[strlen(filep->d_name)-3],".so"))
		continue;

	    if(filename)
		free(filename);

	    filename	= pathadd('/',dirname,filep->d_name,NULL);

	    /* and if we can't open it...  this should at least
	     * give an error message, and, actually, perhaps
	     * even be fatal, but I haven't decided on that yet...
	     * so for now it just skips this file. */
	    if(!(library = dlopen(filename, RTLD_LAZY)))
	    {
		fprintf(stderr,"Warning: couldn't map into %s: %s\n",
			filep->d_name, dlerror());
		continue;
	    }
	    
	    if((*verbosity) > 0)
		/* Let the user know what's going on... */
		fprintf(stderr, "Registering %s/%s...\n",
		    dirname, filep->d_name);

	    /* Now that we have a library, try to find it's
	     * run_registration function. */

	    /* note: some compilers may give a warning about the
	       next line, something like "cast between
	       pointer-to-object and pointer-to-function".  This
	       is safe to ignore...  and an RFE has been filed
	       against at least one UNIX vendor's compilers to
	       fix that. */

	    if(!(run_registration =
		(int (*)(void))dlsym(library,"run_registration")))
	    {
		if((*verbosity) > 0)
		    fprintf(stderr, "FAILED\n");

		fprintf(stderr, "%s: %s",
		    filep->d_name, dlerror());

		/* if it doesn't have one, we won't need it later. */
		if(dlclose(library))
		{
		    /* very unlikely, I suspect... */
		    fprintf(stderr, "Warning: dlclose failed: %s\n",
			dlerror());
		}

		/* (so skip to the next file.) */
		continue;
	    }

	    /* otherwise, leave it open, we'll need it to be
	     * open to call functions from within it. */

	    /* including now, when we ask it to register it's
	     * functions. */
	    (*run_registration)();
	}
    }

    /* don't need to save these... */
    free(dirname);
    free(ldlpath);
}

int	register_task(unsigned long type, unsigned long creator,
			pi_cb callback)
{
    if((verbosity != NULL) && ((*verbosity) > 0))
	printf("Got registration: %.4s/%.4s %p (in slot %d)\n",
	    (char *)&type, (char *)&creator, callback, callback_count);

    callback_count++;

    if(registry_resize())
	return(-1);

    callbacks[callback_count-1].type		= type;
    callbacks[callback_count-1].creator		= creator;
    callbacks[callback_count-1].callback	= callback;

    return(callback_count);
}

int	registry_resize(void)
{
    cb_registration	*tmp;

    if(callback_count < callback_size)
	return(0);
    
    callback_size	+= 100;

    tmp	= realloc(callbacks,sizeof(cb_registration) * callback_size);

    if(!tmp)
	return(-1);

    callbacks		= tmp;

    return(0);
}

int	run_callback(PilotConnection *pilot)
{
    int			i, j, hits;
    unsigned long	t, c, cbt, cbc;
    char		*ptr1, *ptr2;

    t	= pilot->dbinfo.type;
    c	= pilot->dbinfo.creator;

    for(i=0,hits=0;i<callback_count;i++)
    {
	cbt	= callbacks[i].type;
	cbc	= callbacks[i].creator;

	/* convert asterisks to current target to allow wildcards */

	/* first check single asterisks */
	if(cbt == '*')
	    cbt	= t;
	if(cbc == '*')
	    cbc	= c;

	/* then check for embedded asterisks */
	for(j=0;j<4;j++)
	{
	    ptr1	= (char *)&cbt;
	    ptr2	= (char *)&t;
	    if(ptr1[j] == '*')
		ptr1[j] = ptr2[j];

	    ptr1	= (char *)&cbc;
	    ptr2	= (char *)&c;
	    if(ptr1[j] == '*')
		ptr1[j] = ptr2[j];
	}

	if(cbt == t && cbc == c)
	{
	    (callbacks[i].callback)(pilot);
	    hits++;
	}
    }

    return(hits);
}

void	backup_data(PilotConnection *pilot)
{
    int			i, errors;
    char		*name;
    struct stat		fileinfo;

    errors	= 0;

    /* allow us to use verbosity as a global within this file,
       while still allowing possible updates from elsewhere.
       This is likely to be redundant, but that shouldn't hurt.  */
    verbosity	= &(pilot->verbosity);

    send_message("Running backups...\n");

    for(i = 0; dlp_ReadDBList(pilot->connectionfd,0,0x80,
				i,&(pilot->dbinfo)) >= 0;)
    {
	i	= pilot->dbinfo.index + 1;

	if((*verbosity) > 3)
	    send_message("\nin backup_data, i = %d\n",i);

	if((*verbosity) > 3)
	{
	    send_message("db.flags = '0x%x'\n",	pilot->dbinfo.flags);
	    send_message("db.type  = '%.4s'\n",	pilot->dbinfo.type);
	    send_message("db.crtr  = '%.4s'\n",	pilot->dbinfo.creator);
	    send_message("db.name  = '%s'\n",	pilot->dbinfo.name);
	}

	name	= pathadd('.', pilot->dbinfo.name,
			((pilot->dbinfo.flags&dlpDBFlagResource)?
			    "prc":"pdb"), NULL);

	if(stat(name,&fileinfo) < 0)
	{
	    bzero((void *)&fileinfo,sizeof(fileinfo));
	}

	if((*verbosity) > 0)
	{
	    send_message("Database file: %s\n",name);
	    send_message("Last backup: %s",
		ctime(&pilot->dbinfo.moddate));
	}

	run_callback(pilot);

	send_message("Backing up: %s... ",name);

	/* unfortunately, the moddate saving doesn't seem to
	work.  :-(   so, I'm not backing up prc files that exist
	at all...  */
	if(
		/* ( */
		(fileinfo.st_ino != 0)
		/* || (pilot->dbinfo.moddate >= fileinfo.st_mtime)) */
		&& (pilot->dbinfo.flags & dlpDBFlagResource))
	{
	    send_message("Skipped (have up-to-date saved copy)\n");
	    continue;
	}
	else if(pilot->dbinfo.flags & dlpDBFlagResource)
	{
	    /* ought to install new copy of .prc file here */
	}

	pilot->savefile	= pi_file_create(name, &(pilot->dbinfo));
	if(!(pilot->savefile))
	{
	    send_message(
		"FATAL: Unable to open %s, aborting backups.\n", name);
	    errors	= -1;
	    break;
	}

	if(pi_file_retrieve(pilot->savefile,pilot->connectionfd,0)<0)
	{
	    send_message(
		"ERROR: Unable to save %s, continuing.\n",name);
	    errors++;
	}

	pi_file_close(pilot->savefile);
	send_message("Done.\n");
    }

    if(errors < 0)
    {
	send_message("Backups failed!\n");
    }
    else
    {
	send_message("Backups complete (%d error(s) detected)\n",
			errors);
    }
}
